#include <QTimer>
#include "SendBuffer.h"
#include "vmc.h"
#include <QDebug>

SendBuffer::SendBuffer(COM_interface *cinterface, QObject *parent) :
    QObject(parent)
{
    qRegisterMetaType<SEND_ERROR>("SEND_ERROR");
    this->vmc = (VMC*)parent;
    this->interface = cinterface;

    this->responseTimeoutTimer = new QTimer(this);
    this->responseTimeoutTimer->setInterval(1000);      // resend message to vmc after 1s without response
    this->responseTimeoutTimer->setSingleShot(false);

    connect(this->responseTimeoutTimer, SIGNAL(timeout()), this, SLOT(onResponseTimeoutTimerTimeout()));
    responseTimerTimeoutCounter = 0;

    resendCounter = 0;

}


SendBuffer::~SendBuffer()
{
    delete(this->responseTimeoutTimer);
}



/****************************************************************
 *  public interface
 *
 ****************************************************************/

/**
 * from vmc: sendMessage / insert Message in Message-Queue
 *
 * interactiveFlag: (default = true)
 * Interactive commands are commands resulting from user input (e.g. button clicks).
 *      This commands are enqueued only if responseTimeoutTimer is not active.
 *      This ensures, that such messages are sent in sequence and only, if an answer
 *      from vmc was received.
 * Non-interactive commands are data messages which are sent to vmc, caused e.g. by
 * an internal status change.
 *
 */
int SendBuffer::SendMessage(QByteArray & ba, bool enqueue)
{

    // prevent queue for growing endlessly:

    // check size of queue
    // if queue-size is greater than SB_MAX_SENDQUEUESIZE, remove oldest (first) element

    qDebug() << "SendBuffer::SendMessage() messageQueue.size() = " << this->messageQueue.size();

    if (this->messageQueue.size() > SB_MAX_SENDQUEUESIZE) {
        messageQueue.removeFirst();
        this->resendCounter = 0;
    }

    // put message in queue
    if (enqueue)
        this->messageQueue.enqueue(ba);


    /* interactive messages are sent only, if the response timer is not active.
     * this should prevent successive sending of the same message during waiting
     * for an answer from vmc.
     */
    if (! this->responseTimeoutTimer->isActive())  {

        if(!enqueue)
            this->messageQueue.enqueue(ba);

        // try to send message
        this->PrivateSendMessage();
    }

    return 0;
}


/****************************************************************
 *  private interface
 *
 */


/****************************************************************
 * handle ACK/NACK response from vmc
 *
 */
void SendBuffer::onReceiveResponse(quint8 response)
{
    if (this->messageQueue.size() == 0) return;

    // reset counter:
    this->responseTimerTimeoutCounter = 0;

    //if (response == NACK)  qDebug() << "onReceiveResponse() NACK";
    //if (response == ACK)   qDebug() << "onReceiveResponse() ACK";
    qDebug() << "--- onReceiveResponse() messageQueue.size() = " << this->messageQueue.size();



    /* resend command on NACK */
    if (response == NACK) {

        // stop timeoutTimer
        this->responseTimeoutTimer->stop();

        if (this->resendCounter > SB_MAX_RESEND) {
            // command could not be processed on peer: remove it from message queue
            this->messageQueue.removeFirst();
            this->resendCounter = 0;
        }
        else {
            // else: resend command
            if (this->messageQueue.size() != 0) this->PrivateSendMessage();
        }

    }

    /* send next command on ACK */
    if (response == ACK) {

        // stop timeoutTimer
        this->responseTimeoutTimer->stop();

        // remove old message from message-queue
        this->messageQueue.removeFirst();
        this->resendCounter = 0;

        if (this->messageQueue.size() != 0) this->PrivateSendMessage();
    }

}

/* this timeout-handler is called, if no response (ACK/NACK) was received from vmc
 */
void SendBuffer::onResponseTimeoutTimerTimeout()
{

    qDebug() << "onResponseTimeoutTimerTimeout() messageQueue.size() = " << this->messageQueue.size();

    if (this->messageQueue.size() == 0) {
        return;
    }

    // wakeup vmc
    emit this->sendError(SEND_ERROR::RESPONSE);

    if (this->resendCounter > SB_MAX_RESEND) {
       this->messageQueue.removeFirst();
       this->resendCounter = 0;
       this->responseTimerTimeoutCounter++;
       this->responseTimeoutTimer->stop();
       emit this->sendError(SEND_ERROR::RESPONSE_TIMER_TIMEOUT);
       if (this->responseTimerTimeoutCounter > 10) {
            // switch to self out-of-order mode ...
            emit this->sendError(SEND_ERROR::MULTIPLE_RESPONSE_TIMER_TIMEOUT);
       }
       // try to send an possible next message in order
       // to empty queue:
       if (!this->messageQueue.isEmpty()) {
           PrivateSendMessage();
       }
    }

    // send next message
    else {
        PrivateSendMessage();
    }
}


/* send first message in Message-Queue
 *
 * preconditon: at least one message is in queue!
 */
void SendBuffer::PrivateSendMessage()
{
   // take message from queue
   QByteArray ba = this->messageQueue.head();


   // put bytes on serial line
   for (int i = 0; i < ba.size(); i++) {
       this->interface->putChar(ba.at(i));
   }

   // do logging here
   emit this->writeLogSent(ba);


   this->resendCounter++;

   // start timeoutTimer
   this->responseTimeoutTimer->start();


}
