#include "vmc.h"
#include "ATBAPP.h"

#include <QObject>
#include <QProcess>
#include <QTextCodec>
#include <QDateTime>
#include <QTimer>
#include <QDebug>
#include <QSettings>


VMC::VMC(QObject *appControl, QSettings const *settings,
         QObject *parent)
  : QObject(parent) {
    this->flag_blockVMCScreen = 0;
    this->currentCachedScreen = VMC_CMD_SCREEN_START;

    this->m_appControl = appControl;
    this->m_settings = settings;

    this->com_interface = new COM_interface(this);
    this->receiveBuffer = new ReceiveBuffer(com_interface, this);
    this->sendBuffer    = new SendBuffer(com_interface, this);

    connect(receiveBuffer, SIGNAL(ReceiveResponse(quint8)), sendBuffer, SLOT(onReceiveResponse(quint8)));

    com_interface->open(settings->value("VMC/serialPort", "ttymxc2").toString(),
                        settings->value("VMC/baudrate", "115200").toInt());

    connect(receiveBuffer, SIGNAL(ReceiveError(RECEIVE_ERROR)), this, SLOT(onReceiveError(RECEIVE_ERROR)), Qt::QueuedConnection);
    connect(sendBuffer, SIGNAL(sendError(SEND_ERROR)), this, SLOT(onSendError(SEND_ERROR)));

    this->sendDelayTimer = new QTimer;
    this->sendDelayTimer->setSingleShot(true);
    connect(this->sendDelayTimer, SIGNAL(timeout()), this, SLOT(onDelayedMessageTimerTimeout()));

    // send start-command to vmc:
    this->SendSystemMessageStart();
}


VMC::~VMC() {
    delete(this->sendBuffer);
    delete(this->receiveBuffer);
    delete(this->com_interface);
}


/********************************************************************************
 * S E N D    M E S S A G E S      t o   V M C
 *
 *
 */

int VMC::SendMessage(QString msg)
{
    // TODO: use
    // this->SendMessage(msg.toUtf8());

    QByteArray ba;
    this->SendMessage(ba.append(msg));

    return 0;
}


int VMC::SendMessage(QByteArray ba, bool enqueue)
{
    quint8 CRC = 0;


    // calculate CRC
    for (int i = 0; i < ba.size(); i++) {
        CRC += ba.at(i);
    }
    ba.append(CRC);

    this->sendBuffer->SendMessage(ba, enqueue);

    return 0;
}


int VMC::SendACK()
{
    this->com_interface->putChar(ACK);
    emit this->writeLogSent(QByteArray(1, ACK));
    return 0;
}

int VMC::SendNACK()
{
    this->com_interface->putChar(NACK);
    emit this->writeLogSent(QByteArray(1, NACK));
    return 0;
}


int VMC::sendButtonChar(QChar c)
{
    QByteArray msg;

    msg.append(QByteArray::fromHex("3E4531"));
    msg.append((quint8)0x30 + (quint8)c.toLatin1());

    return SendMessage(msg);
}

int VMC::SendButton0()              //  0   -   3E 45 31 30 E4
{
    return SendMessage(QByteArray::fromHex("3E453130"));
}

int VMC::SendButton1()              //  1   -   3E 45 31 31 E5
{
    return SendMessage(QByteArray::fromHex("3E453131"));
}

int VMC::SendButton2()              //  2   -   3E 45 31 32 E6
{
    return SendMessage(QByteArray::fromHex("3E453132"));
}

int VMC::SendButton3()              //  3   -   3E 45 31 33 E7
{
    return SendMessage(QByteArray::fromHex("3E453133"));
}

int VMC::SendButton4()              //  4   -   3E 45 31 34 E8
{
    return SendMessage(QByteArray::fromHex("3E453134"));
}

int VMC::SendButton5()              //  5   -   3E 45 31 35 E9
{
    return SendMessage(QByteArray::fromHex("3E453135"));
}

int VMC::SendButton6()              //  6   -   3E 45 31 36 EA
{
    return SendMessage(QByteArray::fromHex("3E453136"));
}

int VMC::SendButton7()              //  7   -   3E 45 31 37 EB
{
    return SendMessage(QByteArray::fromHex("3E453137"));
}

int VMC::SendButton8()              //  8   -   3E 45 31 38 EC
{
    return SendMessage(QByteArray::fromHex("3E453138"));
}

int VMC::SendButton9()              //  9   -   3E 45 31 39 ED
{
    return SendMessage(QByteArray::fromHex("3E453139"));
}

int VMC::SendButtonNext()           // 10   -   3E 45 31 3A EE
{
    return SendMessage(QByteArray::fromHex("3E45313A"));
}

int VMC::SendButtonBack()           // 11   -   3E 45 31 3B EF
{
    return SendMessage(QByteArray::fromHex("3E45313B"));
}

int VMC::SendButtonCancel()           // 12   -   3E 45 31 3C F0
{
    return SendMessage(QByteArray::fromHex("3E45313C"));
}

int VMC::SendButtonStart()          // 13   -   3E 45 31 3D F1
{
#if defined (ARCH_PTU4) || defined(ARCH_PTU5)
    // Start button wakes vmc:
    emit wakeVMC();
#endif
    return SendMessage(QByteArray::fromHex("3E45313D"), true);
}

int VMC::SendButtonStart2()         //      -   3E 45 31 A0 cs
{
#if defined (ARCH_PTU4) || defined(ARCH_PTU5)
    // Start button wakes vmc:
    emit wakeVMC();
#endif
    return SendMessage(QByteArray::fromHex("3E4531A0"));
}

int VMC::SendButtonStart3()         //      -   3E 45 31 A1 cs
{
#if defined (ARCH_PTU4) || defined(ARCH_PTU5)
    // Start button wakes vmc:
    emit wakeVMC();
#endif
    return SendMessage(QByteArray::fromHex("3E4531A1"));
}

int VMC::SendButtonStart4()         //      -   3E 45 31 A2 cs
{
#if defined (ARCH_PTU4) || defined(ARCH_PTU5)
    // Start button wakes vmc:
    emit wakeVMC();
#endif
    return SendMessage(QByteArray::fromHex("3E4531A2"));
}


int VMC::SendButtonHelp()           // 14   -   3E 45 31 3E F2
{
    return SendMessage(QByteArray::fromHex("3E45313E"));
}

int VMC::SendButtonTarif()          // 15   -   3E 45 31 3F F3
{
    return SendMessage(QByteArray::fromHex("3E45313F"));
}

int VMC::SendButtonArrowLeft()      //      -   3E 45 31 83 cs
{
    return SendMessage(QByteArray::fromHex("3E453183"));
}

int VMC::SendButtonArrowRight()     //      -   3E 45 31 84 cs
{
    return SendMessage(QByteArray::fromHex("3E453184"));
}

int VMC::SendButtonArrowUp()        //      -   3E 45 31 85 cs
{
    return SendMessage(QByteArray::fromHex("3E453185"));
}

int VMC::SendButtonArrowDown()      //      -   3E 45 31 86 cs
{
    return SendMessage(QByteArray::fromHex("3E453186"));
}

int VMC::SendButtonCardTypeLeft()   //      -   3E 45 31 87 cs
{
    return SendMessage(QByteArray::fromHex("3E453187"));
}

int VMC::SendButtonCardTypeRight()  //      -   3E 45 31 88 cs
{
    return SendMessage(QByteArray::fromHex("3E453188"));
}

int VMC::SendButtonA() {              // 'A'  -   3E 45 31 41 F4
    return SendMessage(QByteArray::fromHex("3E453141"));
}

int VMC::SendButtonB() {              // 'B'  -   3E 45 31 42 F5
    return SendMessage(QByteArray::fromHex("3E453142"));
}

int VMC::SendButtonC() {              // 'C'  -   3E 45 31 43 F6
    return SendMessage(QByteArray::fromHex("3E453143"));
}

int VMC::SendButtonD() {              // 'D'  -   3E 45 31 44 F7
    return SendMessage(QByteArray::fromHex("3E453144"));
}

int VMC::SendButtonE() {              // 'E'  -   3E 45 31 45 F8
    return SendMessage(QByteArray::fromHex("3E453145"));
}

int VMC::SendButtonF() {              // 'F'  -   3E 45 31 46 F9
    return SendMessage(QByteArray::fromHex("3E453146"));
}

int VMC::SendButtonG() {              // 'G'  -   3E 45 31 47 FA
    return SendMessage(QByteArray::fromHex("3E453147"));
}

int VMC::SendButtonH() {              // 'H'  -   3E 45 31 48 FB
    return SendMessage(QByteArray::fromHex("3E453148"));
}

int VMC::SendButtonI() {              // 'I'  -   3E 45 31 49 FC
    return SendMessage(QByteArray::fromHex("3E453149"));
}

int VMC::SendButtonJ() {              // 'J'  -   3E 45 31 4A FD
    return SendMessage(QByteArray::fromHex("3E45314A"));
}

int VMC::SendButtonK() {              // 'K'  -   3E 45 31 4B FE
    return SendMessage(QByteArray::fromHex("3E45314B"));
}

int VMC::SendButtonL() {              // 'L'  -   3E 45 31 4C FF
    return SendMessage(QByteArray::fromHex("3E45314C"));
}

int VMC::SendButtonM() {              // 'M'  -   3E 45 31 4D 00
    return SendMessage(QByteArray::fromHex("3E45314D"));
}

int VMC::SendButtonN() {              // 'N'  -   3E 45 31 4E 01
    return SendMessage(QByteArray::fromHex("3E45314E"));
}

int VMC::SendButtonO() {              // 'O'  -   3E 45 31 4F 02
    return SendMessage(QByteArray::fromHex("3E45314F"));
}

int VMC::SendButtonP() {              // 'P'  -   3E 45 31 50 03
    return SendMessage(QByteArray::fromHex("3E453150"));
}

int VMC::SendButtonQ() {              // 'Q'  -   3E 45 31 51 04
    return SendMessage(QByteArray::fromHex("3E453151"));
}

int VMC::SendButtonR() {              // 'R'  -   3E 45 31 52 05
    return SendMessage(QByteArray::fromHex("3E453152"));
}

int VMC::SendButtonS() {              // 'S'  -   3E 45 31 53 06
    return SendMessage(QByteArray::fromHex("3E453153"));
}

int VMC::SendButtonT() {              // 'T'  -   3E 45 31 54 07
    return SendMessage(QByteArray::fromHex("3E453154"));
}

int VMC::SendButtonU() {              // 'U'  -   3E 45 31 55 08
    return SendMessage(QByteArray::fromHex("3E453155"));
}

int VMC::SendButtonV() {              // 'V'  -   3E 45 31 56 09
    return SendMessage(QByteArray::fromHex("3E453156"));
}

int VMC::SendButtonW() {              // 'W'  -   3E 45 31 57 0A
    return SendMessage(QByteArray::fromHex("3E453157"));
}

int VMC::SendButtonX() {              // 'X'  -   3E 45 31 58 0B
    return SendMessage(QByteArray::fromHex("3E453158"));
}

int VMC::SendButtonY() {              // 'Y'  -   3E 45 31 59 0C
    return SendMessage(QByteArray::fromHex("3E453159"));
}

int VMC::SendButtonZ() {              // 'Z'  -   3E 45 31 5A 0D
    return SendMessage(QByteArray::fromHex("3E45315A"));
}

int VMC::SendButtonUE() {            // 'Ü'   -   3E 45 31 DC 90   - ISO 8859-1 (Latin-1): 'Ü'
    return SendMessage(QByteArray::fromHex("3E4531DC"));
}

int VMC::SendButtonOE() {            // 'Ö'   -   3E 45 31 D6 8a   - ISO 8859-1 (Latin-1): 'Ö'
    return SendMessage(QByteArray::fromHex("3E4531D6"));
}

int VMC::SendButtonAE() {            // 'Ä'  -   3E 45 31 C4 78    - ISO 8859-1 (Latin-1): 'Ä'
    return SendMessage(QByteArray::fromHex("3E4531C4"));
}


int VMC::SendButtonDEL() {            // DEL  -   3E 45 31 7F 33
    return SendMessage(QByteArray::fromHex("3E45317F"));
}

int VMC::SendButtonDash() {           // '-'  -   3E 45 31 2D E1
    return SendMessage(QByteArray::fromHex("3E45312D"));
}

int VMC::SendButtonSpace() {          // ' '  -   3E 45 31 20 D4
    return SendMessage(QByteArray::fromHex("3E453120"));
}

int VMC::SendButtonClose() {          //      -   3E 45 31 80 cs
    return SendMessage(QByteArray::fromHex("3E453180"));
}

int VMC::SendButtonYes() {            //      -   3E 45 31 81 cs
    return SendMessage(QByteArray::fromHex("3E453181"));
}

int VMC::SendButtonNo() {             //      -   3E 45 31 82 cs
    return SendMessage(QByteArray::fromHex("3E453182"));
}

int VMC::SendButtonVehicleDefintion1() {   // 0x90 -   3E 45 31 90 cs
    return SendMessage(QByteArray::fromHex("3E453190"));
}

int VMC::SendButtonVehicleDefintion2() {   // 0x91 -   3E 45 31 91 cs
    return SendMessage(QByteArray::fromHex("3E453191"));
}

int VMC::SendButtonVehicleDefintion3() {   // 0x92 -   3E 45 31 92 cs
    return SendMessage(QByteArray::fromHex("3E453192"));
}

int VMC::SendButtonVehicleDefintion4() {   // 0x93 -   3E 45 31 93 cs
    return SendMessage(QByteArray::fromHex("3E453193"));
}

int VMC::SendButtonVehicleDefintion5() {   // 0x94 -   3E 45 31 94 cs
    return SendMessage(QByteArray::fromHex("3E453194"));
}

int VMC::SendButtonVehicleDefintion6() {   // 0x95 -   3E 45 31 95 cs
    return SendMessage(QByteArray::fromHex("3E453195"));
}

int VMC::SendButtonVehicleDefintion7() {   // 0x96 -   3E 45 31 96 cs
    return SendMessage(QByteArray::fromHex("3E453196"));
}

int VMC::SendButtonVehicleDefintion8() {   // 0x97 -   3E 45 31 97 cs
    return SendMessage(QByteArray::fromHex("3E453197"));
}

int VMC::SendButtonPaymethode1() {         // 0x98 -   3E 45 31 98 cs
    return SendMessage(QByteArray::fromHex("3E453198"));
}

int VMC::SendButtonPaymethode2() {         // 0x99 -   3E 45 31 99 cs
    return SendMessage(QByteArray::fromHex("3E453199"));
}

int VMC::SendButtonPaymethode3() {         // 0x9a -   3E 45 31 9a cs
    return SendMessage(QByteArray::fromHex("3E45319a"));
}

int VMC::SendButtonPaymethode4() {         // 0x9b -   3E 45 31 9b cs
    return SendMessage(QByteArray::fromHex("3E45319b"));
}


int VMC::SendSystemMessage(quint8 nr)
{
    QByteArray msg;

    msg.append(QByteArray::fromHex("3EF031"));
    msg.append((quint8)0x30 + (quint8)nr);
    return SendMessage(msg);
}


int VMC::SendSystemMessageStart()
{
    return this->SendSystemMessage(1);
}

int VMC::SendSystemMessageTimeout()
{
    return this->SendSystemMessage(2);
}

int VMC::SendLangSwitch(quint8 nr)
{
    QByteArray msg;

    msg.append(QByteArray::fromHex("3E4631"));
    msg.append((quint8)0x30 + (quint8)nr);
    return SendMessage(msg);
}


/* Send last line of Service Menu to VMC
 *      we send 20 characters
 *
 *      hmi -> vmc: TODO
 */
int VMC::SendButtonServiceOK(QString MessageString)
{
   QString msg = MessageString.leftJustified(20,' ',true);


   QByteArray message(QByteArray::fromHex("3E1014")); // TODO
   message.append(msg.toLatin1());
   SendMessage(message);

   return 1;
}


/* Send Version String to VMC
 *      currently only 20 characters are sent.
 *
 *      vmc -> hmi: 3E 10 31 31 B0
 *      hmi -> vmc: 3E 10 14 56 30 2E 30 32 2E 37 2F 32 30 31 31 30 37 30 34 20 20 20 20 1B
 */
int VMC::SendVersion(QString VersionString) {
   VersionString.truncate(20);
   QByteArray message(QByteArray::fromHex("3E1014"));
   message.append(VersionString.toLatin1());
   SendMessage(message);

   return 1;
}


/* Send Current PTU state
 *      currently only 20 characters are sent.
 *
 *      vmc -> hmi: 3E 11 31 31 B1
 *      hmi -> vmc: 3E 11 32 CM CS cs
 */
int VMC::SendCurrentState()
{
    QByteArray message(QByteArray::fromHex("3E1132"));

    PROGRAM_MODE currentMode;
    if (QMetaObject::invokeMethod(m_appControl,
                                  "getCurrentProgramMode",
                                  Qt::DirectConnection,
                                  Q_RETURN_ARG(PROGRAM_MODE, currentMode))) {
        message.append(static_cast<char>(currentMode));
        SendMessage(message);
    }

    return 1;
}

/* Send lic plate String to VMC
 */
int VMC::SendLicPlate(QString licplate) {   
    QString msg = licplate.leftJustified(9,' ',true);
    QByteArray message(QByteArray::fromHex("3E4739"));
    message.append(msg.toLatin1());
    //message.append(msg.toUtf8()); -> alternative: send Utf8
    SendMessage(message);

    return 1;
}

/* Send generated PIN string to VMC
 */
int VMC::SendGeneratedPinNumber(QString pin)
{
    // DEBUG:
    qDebug() << "SendGeneratedPinNumber(): PIN = " << pin;

    QByteArray message(QByteArray::fromHex("3E4735"));
    message.append(pin.toLatin1());
    SendMessage(message, true);

    return 1;
}


/* Send barcode String to VMC
 *
 * This usese the same interface command like licence plate.
 * Currently, we just transmit more charracters, e.g. 13 chars for EAN-13 code.
 * This can be advanced to configure code settings (e.g. EAN-8, ...)
 *
 * Note: on changing string lenght (e.g. from 13 to 8) one must also adapt
 * the lenght code in message (byte #3 = 0x30 + lenght).
 * E.g. EAN-13: length = 13 characters
 *      => lenght code = 0x30 + 13 = 0x3D
 */
int VMC::SendBarcode(QString barcode) {
    QString msg = barcode.leftJustified(13,' ',true);
    QByteArray message(QByteArray::fromHex("3E473D"));
    message.append(msg.toLatin1());
    SendMessage(message);
    //SendMessageDelayed(message, 300);

    return 1;
}



/**
 * @brief VMC::SendFormatedString
 *          just to send a Latin1-string to vmc.
 * @param formatString
 * @return 1 on success
 */
int VMC::SendFormatedString(QString formatString)
{
    if (formatString.length() == 0) return 0;

    QByteArray message(QByteArray::fromHex("3ED0"));
    message.append((quint8)0x30 + (quint8)formatString.length());
    message.append(formatString.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendLongFormatedStringReference(QString & formatedString)
{
    if (formatedString.length() == 0) return 0;

    QByteArray message(QByteArray::fromHex("3ED1"));
    message.append((quint8)(formatedString.length() & 0x00FF));
    message.append((quint8)((formatedString.length() >> 8) & 0x00FF));
    message.append(formatedString.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendLongFormatedString(QString formatString) {
    if (formatString.length() == 0) return 0;

    QByteArray message(QByteArray::fromHex("3ED1"));
    message.append((quint8)(formatString.length() & 0x00FF));
    message.append((quint8)((formatString.length() >> 8) & 0x00FF));
    message.append(formatString.toLatin1());
    SendMessage(message, true);

    return 1;
}


int VMC::SendFormatedStringError(QString errorString)
{
    if (errorString.length() == 0) return 0;

    QByteArray message(QByteArray::fromHex("3ED0"));
    message.append((quint8)0x30 + (quint8)errorString.length());
    message.append(errorString.toLatin1());
    SendMessage(message, true);

    return 1;
}


/*********************************************************************
 * send bewonerscode data to vmc.
 *
 */
int VMC::SendId(QString id)
{
    QByteArray message(QByteArray::fromHex("3E56"));
    message.append((quint8)0x30 + (quint8)id.length());
    message.append(id.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendCode(QString code)
{
    //  code is filled with ' ' to 20 chars:
    QString msg = code.leftJustified(20,' ',true);
    QByteArray message(QByteArray::fromHex("3E5744"));
    message.append(msg.toLatin1());
    SendMessage(message);

    return 1;
}

/*********************************************************************
 * send ilogs selling data to vmc.
 *
 * TODO: Format of data (e.g. timestamps)
 * * this methods are deprecated *
 * * use SendFormatedString() instead.
 *
 */
int VMC::SendAmountDueTax(QString amount)
{
    QByteArray message(QByteArray::fromHex("3E53"));
    message.append((quint8)0x30 + (quint8)amount.length());
    message.append(amount.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendAmountDueNet(QString amount)
{
    QByteArray message(QByteArray::fromHex("3E51"));
    message.append((quint8)0x30 + (quint8)amount.length());
    message.append(amount.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendAmountDuePeriodStart(QString startTime)
{
    QByteArray message(QByteArray::fromHex("3E54"));
    message.append((quint8)0x30 + (quint8)startTime.length());
    message.append(startTime.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendAmountDuePeriodEnd(QString endTime)
{
    QByteArray message(QByteArray::fromHex("3E50"));
    message.append((quint8)0x30 + (quint8)endTime.length());
    message.append(endTime.toLatin1());
    SendMessage(message, true);

    return 1;
}

int VMC::SendCracePeriod(QString minute)
{
    QByteArray message(QByteArray::fromHex("3E49"));
    message.append((quint8)0x30 + (quint8)minute.length());
    message.append(minute.toLatin1());
    SendMessage(message, true);

    return 1;
}



/*********************************************************************
 * send delayed message
 *
 * note, consecutive calls on this would possibly overwrite an already
 * message, so please use this methode rarely and with care.
 */
int VMC::SendMessageDelayed(QString msg, int delay)
{
    this->delayedMessage = msg;
    this->sendDelayTimer->start(delay);

    return 1;
}
void VMC::onDelayedMessageTimerTimeout()
{
    this->SendMessage(this->delayedMessage);
}






/*********************************************************************
 * get vmc state
 */
quint16 VMC::getCurrentCachedScreen()
{
    return this->currentCachedScreen;
}


/*********************************************************************
 *     H A N D L E      M E S S A G E S    f r o m     V M C
 *
 */


void VMC::onReceiveError(RECEIVE_ERROR error)
{
    switch (error) {
        case RECEIVE_ERROR::CRC:
        case RECEIVE_ERROR::INVALID_COMMAND:
            SendNACK();
            break;
        case RECEIVE_ERROR::NO_COMMAND:
        default: return;

    }
}



void VMC::CommandAvaliable(QByteArray message)
{
    // DEBUG /////////////////////////////////////////////////////
    /*
    QString hexString = "received: ";
    qint64 i;

    for (i=0; i < message.size(); i++) {
        hexString.append((QString(message.at(i))).toLatin1().toHex());
        hexString.append(' ');
    }
    hexString.chop(1);

    qDebug() << "VMC::CommandAvaliable(): message.size(): " << message.size() << "";
    qDebug() << "VMC::CommandAvaliable(): message: "        << hexString      << "";
    */
    //////////////////////////////////////////////////////////////

    emit writeLogReceived(message);

    // now, our message starts with '3e' and has a correct CRC on the end:

    switch ((quint8)message.at(1)) {
        case VMC_CMD_SCREEN:

            // set current cached screen only for certain screens (potential idle screens)
            // NOTE: this has to be checked carefully in customer projects
            switch ((quint8)message.at(3)) {
                case VMC_CMD_SCREEN_START:
                case VMC_CMD_SCREEN_FREEPARK:
                case VMC_CMD_SCREEN_DEFBUTTONS:
                case VMC_CMD_SCREEN_DEFBUTTONS2:
                case VMC_CMD_SCREEN_DEFBUTTONS3:
                     // cache current idle screen
                     this->currentCachedScreen = static_cast<quint16>(message.at(3));
                     break;
                case VMC_CMD_SCREEN_DIAG:
                     // unblock if we go to service mode
                     this->flag_blockVMCScreen = 0;
                     break;
                default:
                     break;
            }
            break;
        case VMC_CMD_TEXT:
            parseTextCommand(message);
            SendACK();
            break;
        case VMC_CMD_TEXT_ENC:
            parseTextCommandEnc(message);
            SendACK();
            break;
        case VMC_CMD_USER_MESSAGE:
            parseUserMessageCommand(message);
            SendACK();
            break;
        case VMC_CMD_SYSTEM:
            parseSystemCommand(message);
            SendACK();
            break;
        case VMC_CMD_MACHINE_NR:
            break;
        case VMC_CMD_PSA_CONFIG:
            parsePSAConfigCommand(message);
            SendACK();
            break;
        case VMC_CMD_LANGUAGE:
            parseLanguageCommand(message);
            SendACK();
            break;
        case VMC_CMD_END_PRG:
            break;
        case VMC_CMD_SW_VERSION:
            SendVersion(ATB_QT_VERSION);
            break;
        case VMC_CMD_STATUS_REQUEST:
            SendCurrentState();
            break;
        case VMC_CMD_DISP_LICENCE_PLATE:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_DISP_PRODUCT:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_DISP_PARKTIME_END:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_DISP_PRICE:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_DATE_TIME:
            parseDateCommand(message);
            SendACK();
            break;
        case VMC_CMD_DISP_LINE1:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_DISP_LINE2:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_DISP_AMOUNT_TO_PAY:
            parseDataDisplayCommand(message);
            SendACK();
            break;
        case VMC_CMD_BUSY:
            parseBusyCommand(message);
            SendACK();
            break;
        case VMC_CMD_MESSAGE_BOX:
            parseShowMessageBoxCommand(message);
            SendACK();
            break;
        case VMC_CMD_SCREEN_CFG:
            parseScreenConfigCommand(message);
            SendACK();
            break;
        case VMC_CMD_VMC_SYSTEM:
            parseVMCSystemMessage(message);
            SendACK();
            break;
        case VMC_CMD_FORMATED_STRING:
            parseVMCFormatedString(message);
            SendACK();
            break;
        default:
            SendNACK();
            break;
    }

    emit CommandProcessed();
}


/* System commands:
 *      vmc -> hmi: 3E 63 32 xx xx cr
 *      hmi -> vmc: ACK/NACK
 *
 * LED-control:
 *      vmc -> hmi: 3E 63 36 xx xx xx xx xx xx cr
 *      hmi -> vmc: ACK/NACK
 */
quint16 VMC::parseSystemCommand(QByteArray &cmd)
{
    quint16 iCommand = 0;

    if( (cmd.at(2) - 0x30) == 0x02) {
       iCommand = (quint8)cmd.at(4) + ((quint16)(cmd.at(3) << 8));
       emit this->sysCommand(iCommand, QByteArray());
    }
    else
    if( (cmd.at(2) - 0x30) == 0x06) {
       iCommand = (quint16)(cmd.at(2) << 8);
       emit this->sysCommand(iCommand, cmd.mid(3,7));
            // note: there is no length check here because
            // cmd is already checked (lenght and checksum)
    }

    return 1;
}



/* Text setting command:
 *      vmc -> hmi: 3E 60 sc tx la lg .. .. ..    .... CRS
 *              sc = screen nr
 *              tx = text nr
 *              la = language
 *              lg = legth
 *      hmi -> vmc: identical message
 */
quint8 VMC::parseTextCommand(QByteArray &cmd)
{
    if (cmd.size() < 7) return 0;

    quint8 screennr = cmd.at(2);
    quint8 textnr   = cmd.at(3);
    quint8 lang     = cmd.at(4)-0x30+1;

    quint8 size     = cmd.at(5)-0x30;

    if (size > cmd.size() - 7) {  size = cmd.size() - 7;  }

    QTextCodec *codec = QTextCodec::codecForName(m_settings->value("CharacterSet", "ISO 8859-15").toByteArray());
    QString text = codec->toUnicode(cmd.mid(6,size));
    qDebug() << "VMC::parseTextCommand() screennr = " << QString::number(screennr, 16);
    qDebug() << "VMC::parseTextCommand() textnr   = " << QString::number(textnr, 16);
    qDebug() << "VMC::parseTextCommand() lang     = " << QString::number(lang, 16);

    //qDebug() << "VMC::parseTextCommand() size     = " << QString::number(size, 16);
    //qDebug() << "VMC::parseTextCommand() text     = " << text;

    // remove whitespace from text start/end except if the text is only one sign!
    // -> this is e.g. for setting one space sign

    return 1;
}






/* Text setting command:
 *      vmc -> hmi: 3E 61 en sc tx la lg .. .. ..    .... CRS
 *              en = encoding
 *              sc = screen nr
 *              tx = text nr
 *              la = language
 *              lg = legth
 *      hmi -> vmc: identical message
 */
quint8 VMC::parseTextCommandEnc(QByteArray &cmd)
{
    if (cmd.size() < 8) return 0;

    quint8 encoding = cmd.at(2);
    quint8 screennr = cmd.at(3);
    quint8 textnr   = cmd.at(4);
    quint8 lang     = cmd.at(5)-0x30+1;

    quint8 size     = cmd.at(6)-0x30;

    if (size > cmd.size() - 8) {  size = cmd.size() - 8;  }

    QByteArray codecName = "ISO 8859-15";

    switch (encoding) {
        case 0x01:  codecName = "ISO 8859-1";    break;
        case 0x02:  codecName = "ISO 8859-2";    break;
        case 0x03:  codecName = "ISO 8859-3";    break;
        case 0x04:  codecName = "ISO 8859-4";    break;
        case 0x05:  codecName = "ISO 8859-5";    break;
        case 0x06:  codecName = "ISO 8859-6";    break;
        case 0x07:  codecName = "ISO 8859-7";    break;
        case 0x08:  codecName = "ISO 8859-8";    break;
        case 0x09:  codecName = "ISO 8859-9";    break;
        case 0x0A:  codecName = "ISO 8859-10";   break;
        case 0x0D:  codecName = "ISO 8859-13";   break;
        case 0x0E:  codecName = "ISO 8859-14";   break;
        case 0x0F:  codecName = "ISO 8859-15";   break;

        case 0x20:  codecName = "Windows-1250";  break;
        case 0x21:  codecName = "Windows-1251";  break;
        case 0x22:  codecName = "Windows-1252";  break;
        case 0x23:  codecName = "Windows-1253";  break;
        case 0x24:  codecName = "Windows-1254";  break;
        case 0x25:  codecName = "Windows-1255";  break;
        case 0x26:  codecName = "Windows-1256";  break;
        case 0x27:  codecName = "Windows-1257";  break;
        case 0x28:  codecName = "Windows-1258";  break;

        case 0x30:  codecName = "KOI8-R";  break;
        case 0x31:  codecName = "KOI8-U";  break;

        case 0xfe:  codecName = "UTF-32";  break;
        case 0xff:  codecName = "UTF-8";  break;

        default:    codecName = "ISO 8859-15"; break;
    }

    QTextCodec *codec = QTextCodec::codecForName(codecName);
    QString text = codec->toUnicode(cmd.mid(7,size));

    qDebug() << "VMC::parseTextCommandEnc() encoding = " << QString(codecName);
    qDebug() << "VMC::parseTextCommandEnc() screennr = " << QString::number(screennr, 16);
    qDebug() << "VMC::parseTextCommandEnc() textnr   = " << QString::number(textnr, 16);
    qDebug() << "VMC::parseTextCommandEnc() lang     = " << QString::number(lang, 16);
    qDebug() << "VMC::parseTextCommandEnc() size     = " << QString::number(size, 16);
    qDebug() << "VMC::parseTextCommandEnc() text     = " << text;

    // remove whitespace from text start/end except if the text is only one sign!
    // -> this is e.g. for setting one space sign

    return 1;
}




/* status data to display:
 *      vmc -> hmi: 3E <dd> <lg> .. .. ..    .... CRS
 *              dd = info data = { 48 | 49 | 50 | 51 | 53 | 54 | 56 }  -> note 52 is command date/time
 *              lg = lenth
 *      hmi -> vmc: identical message
 */
quint8 VMC::parseDataDisplayCommand(QByteArray &cmd)
{
    if (cmd.size() < 4) return 0;

    quint8 infoField = cmd.at(1);
    quint8 size      = cmd.at(2)-0x30;

    if (size > cmd.size() - 4) {  size = cmd.size() - 4;  }

    QTextCodec::ConverterState state;
    QTextCodec *codec = QTextCodec::codecForName("UTF-8");
    //QString text = codec->toUnicode(cmd.mid(3,size));

    // detect whether we got unicode string or VMC legacy format string
    const QByteArray data(cmd.mid(3,size));
    QString text = codec->toUnicode(data.constData(), data.size(), &state);
    if (state.invalidChars > 0) {
        // Not a UTF-8 text
        QTextCodec *defaultCodec = QTextCodec::codecForName(m_settings->value("CharacterSet", "ISO 8859-15").toByteArray());
        text = defaultCodec->toUnicode(data);
    }

    #if defined (USE_PTU_TARIFF)
    // ignore some vmc data commands
    if ( (infoField == 0x51) ||      // amount
         (infoField == 0x50) ||      // parking endtime
         (infoField == 0x48) )       // license plate
    {
        return 0;
    }
    #endif

    if (infoField == 0x56) {    // display amount to pay
        this->privateHandleDataCommand_AmountToPay(infoField, text);
    }

    // Amount ---------------------------------------------------------------------------
    if (infoField == 0x51) {   // display amount
        this->privateHandleDataCommand_Amount(infoField, text);

        #if defined (CUST00309)
        // Neuharlingersiel

        // Workarround for display correct amount in display!
        //   -> text does not contain a '.' or ','

        if ( (text.contains('.')) or
             (text.contains(',')) ) {
                int separatorIndex = text.indexOf("00 ");
                if (separatorIndex >= 0)  {
                    text = text.insert(separatorIndex,',');
                }
                else {
                    text = text.insert(-2,',');
                }
        }
        #endif

    }  // END AMOUNT


    #if defined (CUST00276)
    // Brombachsee -------------------------------------------------------------------------
    if (infoField == 0x50)
    {
        // check format: "hh:mm dd.mm.YYYY"
        QDateTime endTime = QDateTime::fromString(text, "hh:mm dd.MM.yyyy");
        if (endTime.isValid()) {
           // set vendingData
           emit setVendingData("DATETIME_PARKEND", endTime.toString("yyyy-MM-dd HH:mm:ss").toLatin1());

           if (m_settings->value("FeatureShowParkingTime", "false").toBool()) {
               // calculate time difference to current time:
               int secondsTo = QDateTime::currentDateTime().secsTo(endTime);

               int days = (int) secondsTo / 86400;
               int restSeconds =  secondsTo % 86400;

               int hours = (int) restSeconds / 3600;
               restSeconds =  restSeconds % 3600;

               int minutes = (int)  restSeconds / 60;

               // round to full hour: (<- dirty hack!)
               if (minutes > 56) {
                  hours++;
                  minutes = 0;
               }
               // round to full day: (<- dirty hack!)
               if (hours == 23) {
                  days++;
                  hours = 0;
               }

               // TODO: get format string from text file (e.g. screen 0x50 (common))
               //text = QString("%1 Tag(e) %2 Stunde(n) %3 min").arg(days).arg(hours).arg(minutes);

               QString dayText = "";
               if (days == 1) dayText = "1 Tag ";
               if (days > 1)  dayText = QString("%1 Tage ").arg(days);

               QString hourText = "";
               if (hours == 1) hourText = "1 Stunde ";
               if (hours > 1)  hourText = QString("%1 Stunden ").arg(hours);

               QString minuteText = "";
               if (minutes > 0)  minuteText = QString("%1 min").arg(minutes);

               text = dayText + hourText + minuteText;

               if (secondsTo <= 0) { text = "date error"; }
            }    // END: if hasFeatureShowParkingTime
        }        // END: if endTime.isValid
        else {
           text.remove(QChar::Null);
           text = text.trimmed();
        }
    }
    if (infoField == 0x51) {
        text.remove(QChar::Null);
        text = text.simplified();
    }
    // ------------------------------------------------------------------------------------
    #else

    // check format: "hh:mm dd.mm.YYYY"
    QDateTime endTime = QDateTime::fromString(text, "hh:mm dd.MM.yyyy");
    if (endTime.isValid()) {
       // set vendingData
       emit setVendingData("DATETIME_PARKEND", endTime.toString("yyyy-MM-dd HH:mm:ss").toLatin1());
    }

    #endif



    qDebug() << "VMC::parseDataDisplayCommand() infoField = " << QString::number(infoField, 16);

    emit displayData(infoField, text);

    emit retriggerModeSell();

    return 1;
}


/**
 * @brief VMC::privateHandleDataCommand_Amount
 * @param data
 * @param dataString
 * @return
 *
 * Used to set global vendingData "AMOUNT"
 * Note: this should not be used on systems with PTU price calculation because on this
 * systems "AMOUNT" is calculated and managed internally.
 * This method however will possibly orverwrite an internal generated "AMOUNT".
 * Using this method for setting "AMOUNT" is for
 *  - testing / simulation
 *  - for legacy projects
 */
QString & VMC::privateHandleDataCommand_Amount(quint8 data, QString & dataString)
{
    Q_UNUSED(data)

    // check amount string:
    // this is normaly a string: "xxx,yy ZZ"
    // we need amount as a string in cent.

    // DEBUG / SIMULATE card payments - for testing live terminals
    //text = QString::fromUtf8("0,01 \u20AC");

    // trim string
    QString internalAmount = dataString.trimmed();

    // remove currency symbol from string:
    internalAmount = internalAmount.split(' ').first();

    // starts with one or more number chars (digit),
    // followed by 0 or one of the chars ',' or '.'
    // ending with 0,1 or 2 number chars (digits)
    QString pattern = "^\\d+[,\\.]?\\d{0,2}$";


    if(!QRegExp(pattern).exactMatch(internalAmount)) {
         qCritical()  << QDateTime::currentDateTime().toString(Qt::ISODate)
                      << "VMC::parseDataDisplayCommand: amount format error: "
                      << internalAmount;
    }

    // get separator '.' or ','
    if (internalAmount.contains(',')) {
        if (internalAmount.split(',').last().size() == 1) internalAmount.append("0");
        internalAmount.remove(',');
    }
    else
    if (internalAmount.contains('.')) {
        if (internalAmount.split('.').last().size() == 1) internalAmount.append("0");
        internalAmount.remove('.');
    }
    else {
        // in that case, we assume always that amount is in hundreds of units:
        qCritical()  << "  ... consider amount as hundreds of units: " << internalAmount;
    }

    // set vendingData:
    emit setVendingData("AMOUNT", internalAmount.toLatin1());

    return dataString;
}

/**
 * @brief VMC::privateHandleDataCommand_AmountToPay
 * @param data
 * @param dataString
 * @return
 *
 * Used to set global vendingData "PaymentType"
 */
QString & VMC::privateHandleDataCommand_AmountToPay(quint8 data, QString & dataString)
{
    Q_UNUSED(data)

    emit this->setVendingData("PaymentType", QByteArray("0"));   // CoinPayment

    QString pattern1 = "^\\d+$";  // only digits -> hundredth
    QString pattern2 = "^\\d+[,\\.]+\\d{0,2}$";  // digits, separated by ',' or '.'
    QString pattern3 = "^\\d+[,\\.]+\\d{0,2}\\s+\\D+$";  // digits with currency symbol


    if(QRegExp(pattern1).exactMatch(dataString)) {
         //qCritical()  << QDateTime::currentDateTime().toString(Qt::ISODate)
         //             << "VMC::privateHandleDataCommand_AmountToPay: pattern1: "
         //             << dataString;

         // insert decimal separator ','
         dataString.insert(-2,',');

         // append currency symbol
         dataString.append(" ");
         dataString.append(m_settings->value("currencySymbol", "€").toString());
    }
    else
    if(QRegExp(pattern2).exactMatch(dataString)) {
         //qCritical()  << QDateTime::currentDateTime().toString(Qt::ISODate)
         //             << "VMC::privateHandleDataCommand_AmountToPay: pattern2: "
         //             << dataString;

         // fill: two digits behind decimal separator
         if (dataString.contains(',')) {
             if (dataString.split(',').last().size() == 1) dataString.append("0");
             if (dataString.split(',').last().size() == 0) dataString.append("00");
         }
         else
         if (dataString.contains('.')) {
             if (dataString.split('.').last().size() == 1) dataString.append("0");
             if (dataString.split(',').last().size() == 0) dataString.append("00");
         }

         // add currency symbol
         dataString.append(" ");
         dataString.append(m_settings->value("currencySymbol", "€").toString());
    }
    else
    if(QRegExp(pattern3).exactMatch(dataString)) {
         //qCritical()  << QDateTime::currentDateTime().toString(Qt::ISODate)
         //             << "VMC::privateHandleDataCommand_AmountToPay: pattern3: "
         //             << dataString;


    }
    else {
        qCritical()  << QDateTime::currentDateTime().toString(Qt::ISODate)
                              << "VMC::privateHandleDataCommand_AmountToPay: ERROR pattern: "
                              << dataString;
        dataString = "ErrorPattern";
    }

    return dataString;
}






/* date command from vmc: set date and time
 *      vmc -> hmi: 3E 52 36 yy mm dd hh mm ss CRS
 *      hmi -> vmc: 3E 53 36 yy mm dd hh mm ss CRS
 *
 *  Datestring format: YYYY-MM-DD hh:mm[:ss]
 */
quint8 VMC::parseDateCommand(QByteArray &cmd)
{
    if (cmd.size() < 10) return 0;
    if (cmd.at(2) != 0x36) return 0;

    QString yy = QString("20%1").arg((ushort)cmd.at(3),0,10);
    QString mm = QString("%1").arg((ushort)cmd.at(4),0,10);
    QString dd = QString("%1").arg((ushort)cmd.at(5),0,10);
    QString hh = QString("%1").arg((ushort)cmd.at(6),0,10);
    QString mi = QString("%1").arg((ushort)cmd.at(7),0,10);
    QString ss = QString("%1").arg((ushort)cmd.at(8),0,10);

    QString dateTimeString = "\"" + yy + "-"
                           + mm + "-"
                           + dd + " "
                           + hh + ":"
                           + mi + ":"
                           + ss + "\"";


    emit newDateTime(dateTimeString);

    qDebug() << "VMC::parseDateCommand() datetimeCMD: " << dateTimeString;

    return 1;
}


quint8 VMC::parseUserMessageCommand(QByteArray &cmd)
{
    if (cmd.size() < 5) return 0;
    if (cmd.at(2) != 0x32) return 0;

    emit userMessage(cmd.at(3), cmd.at(4));

    return 1;
}


quint8 VMC::parseLanguageCommand(QByteArray &cmd)
{
    if (cmd.size() < 5) return 0;
    if (cmd.at(2) != 0x32) return 0;

    quint8 lang = cmd.at(3);
    quint8 flag = cmd.at(4);

    if (lang > 0x04) return 0;
    if (flag > 0x02) return 0;


    switch (flag) {
        case 0x00:       emit setLanguage(lang);
                         break;
        case 0x01:       emit setLanguage(lang);
                         emit setDefaultLanguage(lang);
                         break;
        case 0x02:       emit setDefaultLanguage(lang);
                         break;
        default:         return 0;
    }

    return 1;
}


quint8 VMC::parseBusyCommand(QByteArray & cmd)
{
    if (cmd.size() < 4) return 0;
    if (cmd.at(2) != 0x31) return 0;

    quint8 flag = cmd.at(3);

    switch (flag) {
        case 0x01:       emit setBusy();
                         break;
        case 0x02:       emit resetBusy();
                         break;
        default:         return 0;
    }

    return 1;
}


quint8 VMC::parseShowMessageBoxCommand(QByteArray & cmd)
{
    if (cmd.size() < 5) return 0;
    if (cmd.at(2) != 0x32) return 0;
    if (cmd.at(4) < 0x30) return 0;

    emit showMessageBox(cmd.at(3), cmd.at(4));

    return 1;
}


quint8 VMC::parseVMCSystemMessage(QByteArray & cmd)
{
    switch (cmd.at(3)) {
        case 0x31: emit this->VMCSystemSellingProcessStart(); break;
        case 0x32: emit this->VMCSystemSellingProcessStop();  break;
        case 0x33: emit this->VMCSystemTransactionSuccess();  break;
        default:
            return 0;
            break;
    }

    return 1;
}

quint8 VMC::parseVMCFormatedString(QByteArray &cmd)
{
    FormatedStringList parameters = cmd.split('#');

    if (parameters.isEmpty()) return 0;

    // remove header
    parameters.takeFirst();
    // remove checksum
    parameters.takeLast();

    emit this->VMCFormatedString(parameters);

    return 1;
}


/*************************************************************************
 * PSA config command has two variants:
 *  1) transmit customer number, group number, zone number and machine number
 *     together in one command, each number represented by an 16-Byte integer.
 *  2) transmit a specified number (= configuration parameter) as a character string.
 *     The transmitted data is identified by a special selector byte.
 *     This variant has some advantages:
 *     - it allows sending arbitrary characters with arbitrary length
 *       -> arbitrary strings can be transmitted
 *     - the command is simple extensible in order to transmit additional data
 *
 *  note: the payload lenght is limited to e.g. 20 characters.
 */
quint8  VMC::parsePSAConfigCommand(QByteArray & cmd)
{

    // check minimum command length
    if (cmd.size() < 5) return 0;

    // payload is at least 1 char and max. 20 char
    if ((cmd.at(2) < 0x32) && (cmd.at(2) > 0x44)) {
        qDebug() << "VMC::parsePSAConfigCommand() - payload length error";
        return 0;
    }

    quint8 payloadLength = cmd.at(2) - 0x30;
    quint8 payloadType   = cmd.at(3);

    // true length check
    if (cmd.size() != payloadLength + 4) {
        qDebug() << "VMC::parsePSAConfigCommand() - command size error";
        return 0;
    }

    QString payload = cmd.mid(4,(payloadLength-1));

    switch (payloadType) {
        case 0x31:      // customer number
            emit setCustNr(payload);
            break;
        case 0x32:      // group number
            emit setGroupNr(payload);
            break;
        case 0x33:      // zone number
            emit setZoneNr(payload);
            break;
        case 0x34:      // machine number
            emit setMachineNr(payload);
            break;
        default:
            qDebug() << "VMC::parsePSAConfigCommand() - unknown payloadType: " << payloadType;
            return 0;
    }

    return 1;
}




void VMC::blockScreenSwitch(quint8 mode)
{
    this->flag_blockVMCScreen = mode;
}



/*********************************************************************
 *     H A N D L E      C C     R E Q U E S T S
 *
 */
void VMC::ccStartTransactionRequest()
{
    #ifdef USE_CC_PLUGIN
    #if defined USE_CC_TRANSACTION
    emit ccStartTransaction();
    #endif
    #endif
}



void VMC::ccStartConfirmTransaction()
{
#ifdef USE_CC_PLUGIN
    emit ccConfirmTransaction();
#endif
}


/**
 * @brief VMC::ccSendReceipt
 *
 * This is used to make a delayed sending to VMC for receipt data
 *
 */
void VMC::ccPrintReceipt(QString receipt)
{
#ifdef USE_CC_PLUGIN
    this->SendLongFormatedString(receipt);
#else
    Q_UNUSED(receipt);
#endif
}


/*********************************************************************
 * customer specific methods
 *
 */

/* for 00318/Nexobility
 *  -> just send button to VMC
 */
void VMC::skipDiscount() {    this->SendButtonPaymethode2();  }



/*********************************************************************
 *     H A N D L E      E R R O R S
 *
 */

/*********************************************************************
 * this is usualy called, if vmc does not send any response after sending
 * a message.
 *
 */
void VMC::onSendError(SEND_ERROR error)
{
    switch (error) {
        case SEND_ERROR::RESPONSE:
            emit wakeVMC();
            break;
        case SEND_ERROR::RESPONSE_TIMER_TIMEOUT:
            // close a possible open busy indicator ...
            // the following does not work like expected: it lasts up to 5s until this methode is called ...
            emit resetBusy();
            break;
        case SEND_ERROR::MULTIPLE_RESPONSE_TIMER_TIMEOUT:
            // TODO: add feature to settings
            if (m_settings->value("FeatureSelf000", "false").toBool()) {
                // Handling on ERROR_RESPONSE_TIMER_TIMEOUT:
                //  -> compare to receive a screen command from vmc to set the ooo-screen:
            }
            break;
        case SEND_ERROR::NO_ERROR:
        default: return;

    }
}
